/*---------------------------------------------------------------------------

    main.c
    
    Copyright (c) 2008 by M. Pernter (iw3aut@yahoo.it)
    
    panel.cpp defines the functions used by the contol panel for interfacing
    between WinRAD and the PMSDR Software Defined Radio using the PIC18F4550 USB
    chip and the CY22394 clock generator and the Si570 clock generator.


    GPL

    This file is part of ExtIO_PMSDR.dll.

    ExtIO_PMSDR.dll is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    ExtIO_PMSDR.dll is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

  ---------------------------------------------------------------------------*/

//---------------------------------------------------------------------------
#include <windows.h>
#include <Dialogs.hpp>
#include <math.h>
#include "main.h"
//#include "communications.h"
#include "codeDeviceManager.h"

#include "panel.h"
#include "about.h"
#include "pmlog.h"
#include "si570.h"
//---------------------------------------------------------------------------

long LOfreq;
unsigned long fREF;
unsigned long data;
unsigned char postdivider, fref_ppm, Q;
unsigned short int P;
unsigned char si570_reg[8];
char USBstring[62];
int USBstring_len;

AnsiString temp;
DeviceManager  deviceManager;

void (* CallBack)(int, int, float, short *);
HWND FindParentHandle();

#pragma argsused

int WINAPI DllEntryPoint(HINSTANCE hinst, unsigned long reason, void* lpReserved)
{
  return 1;
}

void print_freq(void)
{  
   Form1->TEdit_freq_ref->Text= AnsiString().sprintf ("%3.6f MHz",(rfreq*fxtal)/(initial_hsdiv*initial_n1));
   Form1->TEdit_lo_freq->Text= AnsiString().sprintf ("%3.6f MHz",(float)(LOfreq/1000000));
   Form1->TEdit_clka_freq->Text= AnsiString().sprintf ("%3.6f MHz",(float)(4*LOfreq/1000000));

   if (postdivider != 0) // if CY2239x is present on PMSDR-board
        Form1->TEdit_pll_postdivider->Text="/ "+IntToStr(postdivider)+ " (Postdiv.)";
   //Form1->CSpinEdit_ppm->Value=(fref_ppm>>2)&0x3F;
}

void read_si570(bool calc_fxtal)
{
  /* for (char i=0;i<6;i++){
      if(deviceManager.PMSDRReadI2C(I2C_SLA_SI570, i+7, si570_reg[i])) Form1->Memo1->Lines->Add(AnsiString().sprintf ("I2C Read Error"));
      Form1->Memo1->Lines->Add(AnsiString().sprintf ("Si570 Reg.%d = %X",i+7,si570_reg[i]));
      }*/
  //return;
  if(deviceManager.ReadSi570(si570_reg)) Form1->Memo1->Lines->Add(AnsiString().sprintf ("Si570 Read Error"));

  deviceManager.ReadPMSDR(P, Q, fref_ppm, fREF,(unsigned long)(LOfreq), postdivider);
  //Form1->Memo1->Lines->Add(AnsiString().sprintf("Post= %d",postdivider));
  calc_si570registers(si570_reg,calc_fxtal);

  Form1->TEdit_freq_fxtal->Text = AnsiString().sprintf ("%3.6f MHz",fxtal);
  Form1->TEdit_pll_RFREQ->Text = AnsiString().sprintf ("%f",rfreq);
  Form1->TEdit_pll_DCO->Text = AnsiString().sprintf ("%4.5f Mhz",rfreq*fxtal);
  Form1->TEdit_pll_HS_DIV_N->Text = AnsiString().sprintf ("HSDIV=%d N1=%d",initial_hsdiv,initial_n1);
  Form1->TEdit_4xfreq->Text = AnsiString().sprintf ("%3.6f MHz",(rfreq*fxtal)/(initial_hsdiv*initial_n1));
  
  Form1->TEdit_freq_ref->Text= AnsiString().sprintf ("%3.6f MHz",(rfreq*fxtal)/(initial_hsdiv*initial_n1));
  if (postdivider != 0){ // if CY2239x is present on PMSDR-board
        Form1->TEdit_pll_postdivider->Text="/ "+IntToStr(postdivider)+ " (Postdivider)";
        Form1->TEdit_clka_freq->Text= AnsiString().sprintf ("%3.6f MHz",(rfreq*fxtal)/(initial_hsdiv*initial_n1*postdivider));
        Form1->TEdit_lo_freq->Text = AnsiString().sprintf ("%3.6f MHz",(rfreq*fxtal)/(initial_hsdiv*initial_n1*4.0*postdivider));
      }
  else { // CY2239x is NOT present on PMSDR-board
        Form1->TEdit_pll_postdivider->Enabled=false;
        Form1->TEdit_pll_postdivider->Text="Not present";
        Form1->TEdit_clka_freq->Text= AnsiString().sprintf ("%3.6f MHz",(rfreq*fxtal)/(initial_hsdiv*initial_n1));
        Form1->TEdit_lo_freq->Text = AnsiString().sprintf ("%3.6f MHz",(rfreq*fxtal)/(initial_hsdiv*initial_n1*4.0));

      }
  //Form1->MHz->Text = AnsiString().sprintf ("%1.0f",floor((rfreq*fxtal)/(initial_hsdiv*initial_n1*4.0*postdivider)));

}

void calc_si570registers(unsigned char* reg, bool calc_fxtal)
{
// HS_DIV conversion 
	initial_hsdiv = ((reg[0] & 0xE0) >> 5) + 4; // get reg 7 bits 5, 6, 7
	// initial_hsdiv's value could be verified here to ensure that it is one
	// of the valid HS_DIV values from the datasheet. 
	// initial_n1 conversion 
	initial_n1 = (( reg[0] & 0x1F ) << 2 ) + // get reg 7 bits 0 to 4 
	(( reg[1] & 0xC0 ) >> 6 );  // add with reg 8 bits 7 and 8 
	if(initial_n1 == 0)  initial_n1 = 1;
        else if(initial_n1 & 1 != 0) // add one to an odd number
	 initial_n1 = initial_n1 + 1;

	frac_bits = (( reg[2] & 0xF ) * POW_2_24 ); 
	frac_bits = frac_bits + (reg[3] * POW_2_16);
	frac_bits = frac_bits + (reg[4] * 256); 
	frac_bits = frac_bits + reg[5]; 
	
	
	rfreq = frac_bits; 
    rfreq = rfreq / POW_2_28;
    rfreq = rfreq + ( (( reg[1] & 0x3F ) << 4 ) + (( reg[2] & 0xF0 ) >> 4 ) );
    if (calc_fxtal)
        fxtal = (fout0 * initial_n1 * initial_hsdiv) / rfreq; //MHz
    
}

unsigned char SetBits(unsigned char original, unsigned char reset_mask, unsigned 
char new_val) 
{ 
return (( original & reset_mask ) | new_val ); 
} 

int FreqProgSi570(double currentFrequency)
{ 
	int i;
	//float ratio = 0;
	unsigned char counter;
	unsigned char reg137;
	unsigned char buf[7]; 
	unsigned short reg[7];
	unsigned int divider_max;
	unsigned int curr_div; 
	unsigned int whole;
	unsigned char validCombo;
        float curr_n1;
	float n1_tmp;

        reg[6]=1;   // default set the CY2239X Postdivider =1
        currentFrequency*=4.0; // QSD Clock = 4 * frx

        while(currentFrequency < SI570_FOUT_MIN)  // find first currentFrequency above SI570_FOUT_MIN
           {
             reg[6]*=2;       // set the CY2239X Postdivider
             currentFrequency*=2.0;
            }
        postdivider=reg[6];
	// find dividers (get the max and min divider range for the HS_DIV and N1 combo)
	divider_max = floor(FDCO_MAX / currentFrequency); //floorf for SDCC 
	curr_div = ceil(FDCO_MIN / currentFrequency); //ceilf for SDCC 
	validCombo = 0; 
	while (curr_div <= divider_max) 
	{ 
		//check all the HS_DIV values with the next curr_div 
		for(counter=0; counter<6; counter++) 
		{ 
			// get the next possible n1 value 
			hsdiv = HS_DIV[counter]; 
			curr_n1 = (curr_div * 1.0) / (hsdiv * 1.0); 
			// determine if curr_n1 is an integer and an even number or one
			// then it will be a valid divider option for the new frequency 
			n1_tmp = floor(curr_n1); 
			n1_tmp = curr_n1 - n1_tmp; 
			if(n1_tmp == 0.0) 
			{ 
				//then curr_n1 is an integer 
				n1 = (unsigned char) curr_n1; 
				if( (n1 == 1) || ((n1 & 1) == 0) ) 
				{ 
					// then the calculated N1 is either 1 or an even number 
					validCombo = 1; 
				} 
			}
			if(validCombo == 1) break; 
		} 
		if(validCombo == 1) break; 
		//increment curr_div to find the next divider 
		//since the current one was not valid 
		curr_div = curr_div + 1; 
	} 

#ifdef DEBUG	
	printf("FreqProgSi570 -> ");
	printf("validcombo: %d  n1: %x hsdiv: %x fxtal: %f\n",validCombo,n1,hsdiv,fxtal);
#endif
        //Form1->Memo1->Lines->Add(AnsiString().sprintf ("validcombo: %d  n1: %x hsdiv: %x fxtal: %f\n",validCombo,n1,hsdiv,fxtal));

	// if(validCombo == 0) at this point then there's an error 
	// in the calculatio. Check if the provided fout0 and fout1 
	// are not valid frequencies 
	// (old method) new RFREQ calculation -- kept for comparison purposes 
	
	if (validCombo == 0)
		return -1;
	rfreq = (currentFrequency * n1 * hsdiv) / fxtal; //using float
	for(counter = 0; counter < 6; counter++) 
	{
		reg[counter] = 0; //clear registers 
	} 
	
	// new HS_DIV conversion 
	hsdiv = hsdiv - 4; 
	//reset this memory 
	reg[0] = 0; 
	//set the top 3 bits of reg 13 
	reg[0] = (hsdiv << 5); 
	// convert new N1 to the binary representation 
	if(n1 == 1) n1 = 0; 
	else if((n1 & 1) == 0) n1 = n1 - 1; //if n1 is even, subtract one 
	// set reg 7 bits 0 to 4 
	reg[0] = SetBits(reg[0], 0xE0, n1 >> 2); 
	// set reg 8 bits 6 and 7 
	reg[1] = (n1 & 3) << 6; 
	
	// (old method) 
	// convert new RFREQ to the binary representation 
	// separate the integer part 
	whole = floor(rfreq); 
	// get the binary representation of the fractional part 
	frac_bits = floor((rfreq - whole) * POW_2_28); 
	// set reg 12 to 10 making frac_bits smaller by 
	// shifting off the last 8 bits everytime 
	for(counter=5; counter >=3; counter--) 
	{ 
		reg[counter] = frac_bits & 0xFF; 
		frac_bits = frac_bits >> 8;
	} 
	// set the last 4 bits of the fractional portion in reg 9 
	reg[2] = SetBits(reg[2], 0xF0, (frac_bits & 0xF)); 
	// set the integer portion of RFREQ across reg 8 and 9 
	reg[2] = SetBits(reg[2], 0x0F, (whole & 0xF) << 4); 
	reg[1] = SetBits(reg[1], 0xC0, (whole >> 4) & 0x3F); 

        deviceManager.SetSi570(reg, 8); // set Registers on PMSDR-Board
	return 0;
}


void SetFilter(long freq)
{
  if (Form1->TCheck_AutoFilter->Checked)
        {
        freq/=1000000;
        if  (freq >= Form1->F1_1->Value & freq <= Form1->F1_2->Value)Form1->Filterselector->ItemIndex=1;
        else if  (freq >= Form1->F2_1->Value & freq <= Form1->F2_2->Value)Form1->Filterselector->ItemIndex=2;
        else if  (freq >= Form1->F3_1->Value & freq <= Form1->F3_2->Value)Form1->Filterselector->ItemIndex=3;
        else Form1->Filterselector->ItemIndex=0;
        }
}

void ReadFilter(void)
{
  unsigned char reg[6];

        if (deviceManager.ReadFilter(reg)) return;
        Form1->Memo1->Lines->Add("ReadFilter");
        //Form1->Memo1->Lines->Add(AnsiString().sprintf ("%x,%x,%x,%x",reg[0]&0xFF,reg[1]&0xFF,reg[2]&0xFF,reg[3]&0xFF));
        Form1->F1_1->Value= reg[0]&0xFF;
        Form1->F1_2->Value= reg[1]&0xFF;
        Form1->F2_1->Value= reg[2]&0xFF;
        Form1->F2_2->Value= reg[3]&0xFF;
        Form1->F3_1->Value= reg[4]&0xFF;
        Form1->F3_2->Value= reg[5]&0xFF;

}


//---------------------------------------------------------------------------

extern "C"
bool __stdcall __declspec(dllexport) InitHW(char *name, char *model, int& type)
{
  static bool first = true;
  type = DLL_TYPE_NODAC; // 4 ==> data returned via the sound card

  if(first)
  {
    first = false;

    // Initialisation values
    LOfreq = DLL_DEF_FREQ;
    CallBack = NULL;

    //commInit();
  }

  strcpy(name, DLL_DEVICE_NAME);  // change with the name of your HW
  strcpy(model, DLL_DEVICE_TYPE); // change with the model of your HW
  //Form1->Memo1->Lines->Add("InitHW");
  return true;
}

//---------------------------------------------------------------------------
extern "C"
int __stdcall __declspec(dllexport) SetHWLO(long freq)
{

  int returnValue = DLL_FREQ_OK;

  Form1->TEdit_info->Text=IntToStr(freq);

  if (freq < DLL_MIN_FREQ) {
    freq = DLL_MIN_FREQ;
    returnValue = DLL_FREQ_TOLOW; }
  else if (freq > DLL_MAX_FREQ) {
    freq = DLL_MAX_FREQ;
    returnValue = DLL_FREQ_TOHIGH; }

  if (!postdivider & (freq < SI570_MIN_FREQ)) {
    freq = SI570_MIN_FREQ;
    returnValue = DLL_FREQ_TOLOW; }

  FreqProgSi570((double) (freq/1000000.0));
  read_si570(0);
  if (postdivider == 0) // if CY2239x is present on PMSDR-board
        LOfreq = (long) (1000000*(rfreq*fxtal)/(initial_hsdiv*initial_n1*4.0));
  else
        LOfreq = (long) (1000000*(rfreq*fxtal)/(initial_hsdiv*initial_n1*4.0*postdivider));
        
  if (labs(freq - LOfreq)>1 && CallBack) {
  (* CallBack)(-1, 101, 0, NULL);
  }

  if (returnValue == DLL_FREQ_TOLOW && CallBack) {
  (* CallBack)(-1, 101, 0, NULL);
  }

  if (returnValue == DLL_FREQ_TOHIGH && CallBack) {
  (* CallBack)(-1, 101, 0, NULL);
  }

  SetFilter(LOfreq);
  //deviceManager.ReadUSBString(USBstring, USBstring_len);
  //Form1->Memo1->Lines->Add(AnsiString().sprintf ("%s",USBstring));

  return returnValue;
}

//---------------------------------------------------------------------------


extern "C"
bool __stdcall __declspec(dllexport) OpenHW(void)
{
  AnsiString temp;
  
  // Create the control- and aboutbox
  Form1 = new TForm1(NULL);
  Form2 = new TForm2(NULL);

  
  if (  Form1->cmbDEMOCommunicationChannel->Text.SubString(0,3) != "COM")
             deviceManager.SelectDevice(Form1->cmbDEMOCommunicationChannel->Text);

  deviceManager.CloseDevice();
  deviceManager.StopExecution();

  deviceManager.ActiveMode = 1; // Demo Mode
  deviceManager.StartExecution();
  Form1->cmbDEMOCommunicationChannel->Items->Clear();
  Form1->cmbDEMOCommunicationChannel->Items->SetText(deviceManager.GetDeviceList().c_str());
  Form1->cmbDEMOCommunicationChannel->ItemIndex=0;
  deviceManager.SelectDevice(Form1->cmbDEMOCommunicationChannel->Text);
  //ShowStatusMessage("Connecting to board...");
   if (deviceManager.OpenDevice()==false)  {
        MessageDlg("Error on open PMSDR Firmware Version " + deviceManager.ReadVersion(),
               mtError, TMsgDlgButtons() << mbOK, 0);
        return false;
       } 

  Form1->Visible = true;
  Form1->BringToFront();
  Form1->Caption = AnsiString().sprintf("PMSDR V%s", DLL_VERSION);
    
  Form1->Memo1->Lines->Add("PMSDR Firmware Version " + deviceManager.ReadVersion());

  //deviceManager.ResetSi570();
  deviceManager.PMSDRWriteI2C(I2C_SLA_SI570, 135, 0x01);
  //deviceManager.PMSDRWriteI2C(I2C_SLA_SI570, 135, 0x80);
  read_si570(1);
  
  print_freq();

  ReadFilter();

  //Form1->TEdit_info->Text="OpenHW ";
  Form1->Memo1->Lines->Add("OpenHW");

  // Initialisation of the hardware interface
  return true;//return commOpen();
}

//---------------------------------------------------------------------------

extern "C"
int __stdcall __declspec(dllexport) StartHW(long freq)
{
  //long newFreq;
  Form1->TEdit_info->Text="StartHW "+ IntToStr(freq);
  if (freq < DLL_MIN_FREQ)
    freq = DLL_MIN_FREQ;

  if (freq > DLL_MAX_FREQ)
    freq = DLL_MAX_FREQ;


  FreqProgSi570((double) (freq/1000000.0));
  read_si570(0);

  if (postdivider == 0) // if CY2239x is present on PMSDR-board
        LOfreq = (long) (1000000*(rfreq*fxtal)/(initial_hsdiv*initial_n1*4.0));
  else
        LOfreq = (long) (1000000*(rfreq*fxtal)/(initial_hsdiv*initial_n1*4.0*postdivider));


  if (labs(freq - LOfreq)>1 && CallBack) {
    (* CallBack)(-1, 101, 0, NULL);
  }
  Form1->Memo1->Lines->Add("StartHW");
  return true; // number of complex elements returned each
}

//---------------------------------------------------------------------------

extern "C"
void __stdcall __declspec(dllexport) IFLimitsChanged(long low, long high)
{
  Form1->TEdit_info->Text="Limits";      
  Form1->Memo1->Lines->Add(AnsiString().sprintf ("IFLimitsChanged %d %d",low,high));
  return;
}

//---------------------------------------------------------------------------

extern "C"
int __stdcall __declspec(dllexport) GetStatus(void)
{
  Form1->TEdit_info->Text="GetStatus ";
  return true; // status not supported by this specific HW,
}

//---------------------------------------------------------------------------

extern "C"
void __stdcall __declspec(dllexport) StopHW(void)
{
  Form1->TEdit_info->Text="StopHW";
  return; // nothing to do with this specific HW
}

//---------------------------------------------------------------------------

extern "C"
void __stdcall __declspec(dllexport) CloseHW(void)
{
  // Close the USB hardware
  deviceManager.CloseDevice();
  deviceManager.StopExecution();

  // Remove the forms
  delete Form1;
  delete Form2;
  delete FormPMLog;

  return;
}

//---------------------------------------------------------------------------


extern "C"
long __stdcall __declspec(dllexport) GetHWLO(void)
{

      read_si570(0);
      if (postdivider == 0) // if CY2239x is not present on PMSDR-board
        LOfreq = (long) (1000000*(rfreq*fxtal)/(initial_hsdiv*initial_n1*4.0));
  else
        LOfreq = (long) (1000000*(rfreq*fxtal)/(initial_hsdiv*initial_n1*4.0*postdivider));

      //LOfreq = (long) (1000000*(rfreq*fxtal)/(initial_hsdiv*initial_n1*4.0));

  return (long)(LOfreq);
}

//---------------------------------------------------------------------------

extern "C"
long __stdcall __declspec(dllexport) GetHWSR(void)
{
  Form1->TEdit_info->Text="TuneChanged ";
  return DLL_SR_NONE; // this DLL controls just an oscillator, not a digitizer
}

//---------------------------------------------------------------------------

extern "C"
void __stdcall __declspec(dllexport) TuneChanged(long freq)
{
  Form1->TEdit_info->Text="TuneChanged ";
  if (FormPMLog != NULL) {
     FormPMLog->EditFrequency->Text=AnsiString().sprintf("%d,%03d kHz",freq/1000,freq-(freq/1000)*1000 );
     FormPMLog->EditDate->Text=Date();
     FormPMLog->EditTime->Text=Time();
     //ormPMLog->EditFrequency->Text.T
     }
  return;
}

//---------------------------------------------------------------------------

extern "C"
void __stdcall __declspec(dllexport) SetCallback(void (* myCallback)(int, int, float, short *))
{
  CallBack = myCallback;
  return;
}

//---------------------------------------------------------------------------

extern "C"
void __stdcall __declspec(dllexport) ShowGUI(void)
{
  Form1->Visible = true;
  Form1->BringToFront();

  return;
}

//---------------------------------------------------------------------------

extern "C"
void __stdcall __declspec(dllexport) HideGUI(void)
{
  Form1->Visible = false;

  return;
}

//---------------------------------------------------------------------------

extern "C"
void __stdcall __declspec(dllexport) RawDataReady(long samprate, int *Ldata, int *Rdata, int numsamples)
{
  return;
}

//---------------------------------------------------------------------------

extern "C"
void __stdcall __declspec(dllexport) ModeChanged(char mode)
{
  if (Form1 != NULL) Form1->TEdit_info->Text="ModeChanged ";
  if (FormPMLog != NULL) {
     switch (mode)
        {
          case 'C': FormPMLog->ComboBoxMode->ItemIndex=0; break;
          case 'A': FormPMLog->ComboBoxMode->ItemIndex=1; break;
          case 'U': FormPMLog->ComboBoxMode->ItemIndex=2; break;
          case 'L': FormPMLog->ComboBoxMode->ItemIndex=3; break;
          case 'F': FormPMLog->ComboBoxMode->ItemIndex=4; break;
          case 'D': FormPMLog->ComboBoxMode->ItemIndex=5; break;
          case 'E': FormPMLog->ComboBoxMode->ItemIndex=6; break;
        }
     FormPMLog->EditDate->Text=Date();
     FormPMLog->EditTime->Text=Time();
     }
  return;
}

//---------------------------------------------------------------------------

extern "C"
char __stdcall __declspec(dllexport) GetMode(void)
{
   AnsiString t;

   if (FormPMLog != NULL) {
        AnsiString t = FormPMLog->ComboBoxMode->Text;
        Form1->Memo1->Lines->Add("GetMode: "+ t);
        return (char)(t[1]);
      }
   return 'A';
}

//---------------------------------------------------------------------------

extern "C"
long __stdcall __declspec(dllexport) GetTune(void)
{
   AnsiString t;

   if (FormPMLog != NULL) {
        t = FormPMLog->EditFrequency->Text;
        t=t.SubString(1,t.Pos(' ')-1);
        t.Trim();
        t.Delete(t.Pos(','),1);
        t.TrimRight();
        long freq=StrToInt(t);
        Form1->Memo1->Lines->Add("GetTune: "+ t + "Hz");
        return (long)(StrToInt(t));
      }
   return DLL_DEF_FREQ;
}

//---------------------------------------------------------------------------

HWND FindParentHandle()
{
  HWND hwndNext, hWindow = NULL;
  DWORD dwID, dwIDWe;

  /* find window */
  dwIDWe = GetCurrentProcessId();
  hwndNext = GetWindow(GetDesktopWindow(), GW_CHILD);
  while (hwndNext != NULL)
  {
    GetWindowThreadProcessId(hwndNext, &dwID);

    if( dwID == dwIDWe )
    {
      hWindow = hwndNext;
      break;
    }

    hwndNext = GetWindow(hwndNext, GW_HWNDNEXT);
  }

  /* find parent */
  if( hWindow )
  {
    hwndNext = hWindow;
    while( hWindow )
      hWindow = GetWindow(hWindow, GW_OWNER);

    if( !hWindow )
      hWindow = hwndNext;
  }

  //MessageDlg(hWindow->Name, mtInformation, TMsgDlgButtons() << mbOK, 0);

  return hWindow;
}

//---------------------------------------------------------------------------
